/*
 * Asterisk -- A telephony toolkit for Linux.
 *
 * Pickup, channel independent call pickup
 * 
 * Copyright (C) 2004, Junghanns.NET GmbH
 *
 * Klaus-Peter Junghanns <kpj@junghanns.net>
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License
 */

#include <asterisk/lock.h>
#include <asterisk/file.h>
#include <asterisk/logger.h>
#include <asterisk/channel.h>
#include <asterisk/channel_pvt.h>
#include <asterisk/pbx.h>
#include <asterisk/module.h>
#include <asterisk/features.h>
#include <asterisk/options.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <signal.h>

#include <pthread.h>


static char *tdesc = "PickUp/PickDown/Steal";

static char *app = "PickUp";

static char *synopsis = "Channel independent call pickup.";

static char *descrip = 
"  PickDown([group]):  Tries to pickup the first ringing channel with callgroup == group.\n"
"                    If called without the group argument, the pickupgroup of the channel will be used.\n";

static char *app2 = "Steal";

static char *synopsis2 = "Channel independent call stealing. Just like pickup but for answered channels.";

static char *descrip2 = 
"  Steal([group]):  Tries to steal the first bridged channel with callgroup == group.\n"
"                    If called without the group argument, the pickupgroup of the channel will be used.\n";

static char *app3 = "PickDown";

static char *synopsis3 = "Channel independent call pickdown.";

static char *descrip3 = 
"  PickDown([group]):  Tries to hangup the first ringing channel with callgroup == group.\n"
"                    If called without the group argument, the pickupgroup of the channel will be used.\n";

STANDARD_LOCAL_USER;

LOCAL_USER_DECL;

static int my_pickup_call(struct ast_channel *chan, unsigned int pickupgroup, int chanstate, int bridge) {
	struct ast_channel *cur;
	int res = -1;
	cur = ast_channel_walk_locked(NULL);
	while(cur) {
		if ((cur != chan) &&
			(pickupgroup & cur->callgroup) &&
			 (cur->_state == chanstate)) {
			 	break;
		}
		ast_mutex_unlock(&cur->lock);
		cur = ast_channel_walk_locked(cur);
	}
	if (cur) {
		if(option_verbose > 2) {
		    if (chanstate == AST_STATE_RINGING) {
			if (bridge == 1) {
		    	    ast_verbose(VERBOSE_PREFIX_3 "Channel %s picked up ringing channel %s\n",chan->name,cur->name);
			} else {
		    	    ast_verbose(VERBOSE_PREFIX_3 "Channel %s hung up ringing channel %s\n",chan->name,cur->name);
			}
		    } else {
		        ast_verbose(VERBOSE_PREFIX_3 "Channel %s stole channel %s\n",chan->name,cur->name);
		    }
		}
		if (bridge == 1) {
		    if (ast_channel_masquerade(cur, chan)) {
			ast_log(LOG_ERROR, "unable to masquerade\n");
		    }
		    ast_mutex_unlock(&cur->lock);
		    ast_mutex_unlock(&chan->lock);
		    ast_answer(cur);
		} else {
		    cur->_softhangup = AST_SOFTHANGUP_DEV;
		    ast_mutex_unlock(&cur->lock);
		}
    	} else	{
		if(option_verbose > 2) {
		    ast_verbose(VERBOSE_PREFIX_3 "No channel found %d.\n",pickupgroup);
		}
	} 
	return res;
}


static int pickup_exec(struct ast_channel *chan, void *data)
{
	int res=0;
	unsigned int pickupgroup=0;
	struct localuser *u;
	if (!data || !strlen(data)) {
	    pickupgroup = chan->pickupgroup;
	} else {
	    pickupgroup = ast_get_group(data);
	}
	LOCAL_USER_ADD(u);
	if (!res) {
		res = my_pickup_call(chan, pickupgroup, AST_STATE_RINGING, 1);
	}
	if (res > 0)
		res = 0;
	LOCAL_USER_REMOVE(u);
	return res;
}

static int steal_exec(struct ast_channel *chan, void *data)
{
	int res=0;
	unsigned int pickupgroup=0;
	struct localuser *u;
	if (!data || !strlen(data)) {
	    pickupgroup = chan->pickupgroup;
	} else {
	    pickupgroup = ast_get_group(data);
	}
	LOCAL_USER_ADD(u);
	if (!res) {
		res = my_pickup_call(chan, pickupgroup, AST_STATE_UP, 1);
	}
	if (res > 0)
		res = 0;
	LOCAL_USER_REMOVE(u);
	return res;
}

static int pickdown_exec(struct ast_channel *chan, void *data)
{
	int res=0;
	unsigned int pickupgroup=0;
	struct localuser *u;
	if (!data || !strlen(data)) {
	    pickupgroup = chan->pickupgroup;
	} else {
	    pickupgroup = ast_get_group(data);
	}
	LOCAL_USER_ADD(u);
	if (!res) {
		res = my_pickup_call(chan, pickupgroup, AST_STATE_RINGING, 0);
	}
	if (res > 0)
		res = 0;
	LOCAL_USER_REMOVE(u);
	return res;
}

int unload_module(void)
{
	STANDARD_HANGUP_LOCALUSERS;
	ast_unregister_application(app3);
	ast_unregister_application(app2);
	return ast_unregister_application(app);
}

int load_module(void)
{
	ast_register_application(app3, pickdown_exec, synopsis3, descrip3);
	ast_register_application(app2, steal_exec, synopsis2, descrip2);
	return ast_register_application(app, pickup_exec, synopsis, descrip);
}

char *description(void)
{
	return tdesc;
}

int usecount(void)
{
	int res;
	STANDARD_USECOUNT(res);
	return res;
}

char *key()
{
	return ASTERISK_GPL_KEY;
}
